from collections.abc import Sequence
from typing import TypeAlias, TypeVar

from pydantic import BaseModel
from pydantic.json_schema import SkipJsonSchema

from .content import ImageData

AgentID: TypeAlias = str


class AgentPayload(BaseModel):
    # TODO: do we need conversation?
    selected_recipient_ids: SkipJsonSchema[Sequence[AgentID] | None] = None


class AgentState(BaseModel):
    pass


InT = TypeVar("InT", bound=AgentPayload, contravariant=True)  # noqa: PLC0105
OutT = TypeVar("OutT", bound=AgentPayload, covariant=True)  # noqa: PLC0105
StateT = TypeVar("StateT", bound=AgentState, covariant=True)  # noqa: PLC0105


class LLMPromptArgs(BaseModel):
    pass


LLMPrompt: TypeAlias = str
LLMFormattedSystemArgs: TypeAlias = dict[str, str]
LLMFormattedArgs: TypeAlias = dict[str, str | ImageData]



from typing import Any

from pydantic import BaseModel

from ..typing.tool import BaseTool, ToolChoice
from . import (
    ChatCompletionFunctionDefinition,
    ChatCompletionNamedToolChoiceFunction,
    ChatCompletionNamedToolChoiceParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)


def to_api_tool(
    tool: BaseTool[BaseModel, BaseModel, Any],
) -> ChatCompletionToolParam:
    return ChatCompletionToolParam(
        type="function",
        function=ChatCompletionFunctionDefinition(
            name=tool.name,
            description=tool.description,
            parameters=tool.in_schema.model_json_schema(),
            strict=tool.strict,
        ),
    )


def to_api_tool_choice(
    tool_choice: ToolChoice,
) -> ChatCompletionToolChoiceOptionParam:
    if isinstance(tool_choice, BaseTool):
        return ChatCompletionNamedToolChoiceParam(
            type="function",
            function=ChatCompletionNamedToolChoiceFunction(name=tool_choice.name),
        )
    return tool_choice



import ast
import asyncio
from datetime import datetime
import functools
import json
import re
from collections.abc import Callable, Coroutine
from copy import deepcopy
from pathlib import Path
from typing import Any, TypeVar, cast

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo
from tqdm.autonotebook import tqdm


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str:
    """Reads and returns contents of file"""
    try:
        if binary_mode:
            with open(file_path, "rb") as file:
                return file.read()
        else:
            with open(file_path) as file:
                return file.read()
    except FileNotFoundError:
        print(f"File {file_path} not found.")
        return ""


async def asyncio_gather_with_pbar(
    *corouts: Coroutine[Any, Any, Any],
    no_tqdm: bool = False,
    desc: str | None = None,
) -> list[Any]:
    pbar = tqdm(total=len(corouts), desc=desc, disable=no_tqdm)

    async def run_and_update(coro: Coroutine[Any, Any, Any]) -> Any:
        result = await coro
        pbar.update(1)
        return result

    wrapped_tasks = [run_and_update(c) for c in corouts]
    results = await asyncio.gather(*wrapped_tasks)
    pbar.close()

    return results


def get_timestamp() -> str:
    return datetime.now().strftime("%Y%m%d_%H%M%S")


def read_txt(file_path: str) -> str:
    return Path(file_path).read_text()


def format_json_string(text: str) -> str:
    decoder = json.JSONDecoder()
    text = text.replace("\n", "")
    length = len(text)
    i = 0
    while i < length:
        ch = text[i]
        if ch in "{[":
            try:
                _, end = decoder.raw_decode(text[i:])
                return text[i : i + end]
            except ValueError:
                pass
        i += 1

    return ""


def read_json_string(json_str: str) -> dict[str, Any] | list[Any]:
    try:
        json_response = ast.literal_eval(json_str)
    except (ValueError, SyntaxError):
        try:
            json_response = json.loads(json_str)
        except json.JSONDecodeError as exc:
            raise ValueError(
                "Invalid JSON - Both ast.literal_eval and json.loads "
                f"failed to parse the following response:\n{json_str}"
            ) from exc

    return json_response  # type: ignore


def extract_json(json_str: str) -> dict[str, Any] | list[Any]:
    return read_json_string(format_json_string(json_str))


def extract_xml_list(text: str) -> list[str]:
    pattern = re.compile(r"<(chunk_\d+)>(.*?)</\1>", re.DOTALL)

    chunks: list[str] = []
    for match in pattern.finditer(text):
        content = match.group(2).strip()
        chunks.append(content)
    return chunks


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        prompt = (
            read_contents_from_file(prompt_path) if prompt_path is not None else None
        )
    else:
        prompt = prompt_text

    return prompt


def merge_pydantic_models(*models: type[BaseModel]) -> type[BaseModel]:
    fields_dict: dict[str, FieldInfo] = {}
    for model in models:
        for field_name, field_info in model.model_fields.items():
            if field_name in fields_dict:
                raise ValueError(
                    f"Field conflict detected: '{field_name}' exists in multiple models"
                )
            fields_dict[field_name] = field_info

    return create_model("MergedModel", __module__=__name__, **fields_dict)  # type: ignore


def filter_fields(data: dict[str, Any], model: type[BaseModel]) -> dict[str, Any]:
    return {key: data[key] for key in model.model_fields if key in data}


T = TypeVar("T", bound=Callable[..., Any])


def forbid_state_change(method: T) -> T:
    @functools.wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        before = deepcopy(self.__dict__)
        result = method(self, *args, **kwargs)
        after = self.__dict__
        if before != after:
            raise RuntimeError(
                f"Method '{method.__name__}' modified the instance state."
            )
        return result

    return cast("T", wrapper)



from collections.abc import AsyncIterator, Iterable
from typing import Any

from pydantic import BaseModel

from ..typing.completion import Completion, CompletionChunk
from ..typing.content import Content
from ..typing.converters import Converters
from ..typing.message import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from ..typing.tool import BaseTool, ToolChoice
from . import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionAsyncStream,  # type: ignore[import]
    ChatCompletionChunk,
    ChatCompletionContentPartParam,
    # ChatCompletionDeveloperMessageParam,
    ChatCompletionMessage,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
    ChatCompletionUsage,
    ChatCompletionUserMessageParam,
)
from .completion_converters import (
    from_api_completion,
    from_api_completion_chunk,
    from_api_completion_chunk_iterator,
    to_api_completion,
)
from .content_converters import from_api_content, to_api_content
from .message_converters import (
    from_api_assistant_message,
    from_api_system_message,
    from_api_tool_message,
    from_api_user_message,
    to_api_assistant_message,
    to_api_system_message,
    to_api_tool_message,
    to_api_user_message,
)
from .tool_converters import to_api_tool, to_api_tool_choice


class OpenAIConverters(Converters):
    @staticmethod
    def to_system_message(
        system_message: SystemMessage, **kwargs: Any
    ) -> ChatCompletionSystemMessageParam:
        return to_api_system_message(system_message, **kwargs)

    @staticmethod
    def from_system_message(
        raw_message: ChatCompletionSystemMessageParam,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> SystemMessage:
        return from_api_system_message(raw_message, model_id=model_id, **kwargs)

    @staticmethod
    def to_user_message(
        user_message: UserMessage, **kwargs: Any
    ) -> ChatCompletionUserMessageParam:
        return to_api_user_message(user_message, **kwargs)

    @staticmethod
    def from_user_message(
        raw_message: ChatCompletionUserMessageParam,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> UserMessage:
        return from_api_user_message(raw_message, model_id=model_id, **kwargs)

    @staticmethod
    def to_assistant_message(
        assistant_message: AssistantMessage, **kwargs: Any
    ) -> ChatCompletionAssistantMessageParam:
        return to_api_assistant_message(assistant_message, **kwargs)

    @staticmethod
    def from_assistant_message(
        raw_message: ChatCompletionMessage,
        raw_usage: ChatCompletionUsage,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> AssistantMessage:
        return from_api_assistant_message(
            raw_message, raw_usage, model_id=model_id, **kwargs
        )

    @staticmethod
    def to_tool_message(
        tool_message: ToolMessage, **kwargs: Any
    ) -> ChatCompletionToolMessageParam:
        return to_api_tool_message(tool_message, **kwargs)

    @staticmethod
    def from_tool_message(
        raw_message: ChatCompletionToolMessageParam,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> ToolMessage:
        return from_api_tool_message(raw_message, model_id=model_id, **kwargs)

    @staticmethod
    def to_tool(
        tool: BaseTool[BaseModel, BaseModel, Any], **kwargs: Any
    ) -> ChatCompletionToolParam:
        return to_api_tool(tool, **kwargs)

    @staticmethod
    def to_tool_choice(
        tool_choice: ToolChoice, **kwargs: Any
    ) -> ChatCompletionToolChoiceOptionParam:
        return to_api_tool_choice(tool_choice, **kwargs)

    @staticmethod
    def to_content(
        content: Content, **kwargs: Any
    ) -> Iterable[ChatCompletionContentPartParam]:
        return to_api_content(content, **kwargs)

    @staticmethod
    def from_content(
        raw_content: str | Iterable[ChatCompletionContentPartParam],
        **kwargs: Any,
    ) -> Content:
        return from_api_content(raw_content, **kwargs)

    @staticmethod
    def to_completion(completion: Completion, **kwargs: Any) -> ChatCompletion:
        return to_api_completion(completion, **kwargs)

    @staticmethod
    def from_completion(
        raw_completion: ChatCompletion,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> Completion:
        return from_api_completion(raw_completion, model_id=model_id, **kwargs)

    @staticmethod
    def to_completion_chunk(
        chunk: CompletionChunk, **kwargs: Any
    ) -> ChatCompletionChunk:
        raise NotImplementedError

    @staticmethod
    def from_completion_chunk(
        raw_chunk: ChatCompletionChunk,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> CompletionChunk:
        return from_api_completion_chunk(raw_chunk, model_id=model_id, **kwargs)

    @staticmethod
    def from_completion_chunk_iterator(  # type: ignore[override]
        raw_chunk_iterator: ChatCompletionAsyncStream[ChatCompletionChunk],
        model_id: str | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[CompletionChunk]:
        return from_api_completion_chunk_iterator(
            raw_chunk_iterator, model_id=model_id, **kwargs
        )



from collections.abc import Sequence
from typing import Any, Generic, cast, final

from ..agent_message_pool import AgentMessage, AgentMessagePool
from ..comm_agent import CommunicatingAgent
from ..run_context import CtxT, RunContextWrapper
from ..typing.io import AgentID, AgentPayload, AgentState, InT, OutT
from .workflow_agent import WorkflowAgent


class SequentialWorkflowAgent(WorkflowAgent[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    def __init__(
        self,
        agent_id: AgentID,
        subagents: Sequence[
            CommunicatingAgent[AgentPayload, AgentPayload, AgentState, CtxT]
        ],
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
        dynamic_routing: bool = False,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        super().__init__(
            subagents=subagents,
            start_agent=subagents[0],
            end_agent=subagents[-1],  # type: ignore[assignment]
            agent_id=agent_id,
            message_pool=message_pool,
            recipient_ids=recipient_ids,
            dynamic_routing=dynamic_routing,
        )

    @final
    async def run(
        self,
        inp_items: Any | None = None,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        rcv_message: AgentMessage[InT, AgentState] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **kwargs: Any,
    ) -> AgentMessage[OutT, AgentState]:
        agent_message = rcv_message
        for subagent in self.subagents:
            agent_message = await subagent.run(
                inp_items=inp_items,
                rcv_message=agent_message,
                entry_point=entry_point,
                forbid_state_change=forbid_state_change,
                ctx=ctx,
                **kwargs,
            )
            inp_items = None
            entry_point = False

        return cast("AgentMessage[OutT, AgentState]", agent_message)



from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from typing import (
    Any,
    Concatenate,
    ParamSpec,
    TypeAlias,
    TypeVar,
)

MAX_RPM = 1e10


@dataclass
class RateLimiterState:
    next_request_time: float = 0.0


QueryT = TypeVar("QueryT")
QueryR = TypeVar("QueryR")
QueryP = ParamSpec("QueryP")

RetrievalFuncSingle: TypeAlias = Callable[
    Concatenate[QueryT, QueryP], Coroutine[Any, Any, QueryR]
]
RetrievalFuncList: TypeAlias = Callable[
    Concatenate[list[QueryT], QueryP], Coroutine[Any, Any, list[QueryR]]
]

RetrievalMethodSingle: TypeAlias = Callable[
    Concatenate[Any, QueryT, QueryP], Coroutine[Any, Any, QueryR]
]
RetrievalMethodList: TypeAlias = Callable[
    Concatenate[Any, list[QueryT], QueryP], Coroutine[Any, Any, list[QueryR]]
]

RetrievalCallableSingle: TypeAlias = (
    RetrievalFuncSingle[QueryT, QueryP, QueryR]
    | RetrievalMethodSingle[QueryT, QueryP, QueryR]
)

RetrievalCallableList: TypeAlias = (
    RetrievalFuncList[QueryT, QueryP, QueryR]
    | RetrievalMethodList[QueryT, QueryP, QueryR]
)


RateLimDecoratorWithArgsSingle = Callable[
    [RetrievalCallableSingle[QueryT, QueryP, QueryR]],
    RetrievalCallableSingle[QueryT, QueryP, QueryR],
]


RateLimDecoratorWithArgsList = Callable[
    [RetrievalCallableList[QueryT, QueryP, QueryR]],
    RetrievalCallableList[QueryT, QueryP, QueryR],
]



from abc import ABC, abstractmethod
from typing import Any, Generic, Protocol

from pydantic import BaseModel

from .run_context import CtxT, RunContextWrapper
from .typing.io import AgentID, AgentPayload, OutT, StateT
from .typing.tool import BaseTool


class ParseOutputHandler(Protocol[OutT, CtxT]):
    def __call__(
        self, *args: Any, ctx: RunContextWrapper[CtxT] | None, **kwargs: Any
    ) -> OutT: ...


class BaseAgent(ABC, Generic[OutT, StateT, CtxT]):
    @abstractmethod
    def __init__(
        self,
        agent_id: AgentID,
        *,
        out_schema: type[OutT] = AgentPayload,
        **kwargs: Any,
    ) -> None:
        self._state: StateT
        self._agent_id = agent_id
        self._out_schema = out_schema
        self._parse_output_impl: ParseOutputHandler[OutT, CtxT] | None = None

    def parse_output_handler(
        self, func: ParseOutputHandler[OutT, CtxT]
    ) -> ParseOutputHandler[OutT, CtxT]:
        self._parse_output_impl = func

        return func

    def _parse_output(
        self, *args: Any, ctx: RunContextWrapper[CtxT] | None = None, **kwargs: Any
    ) -> OutT:
        if self._parse_output_impl:
            return self._parse_output_impl(*args, ctx=ctx, **kwargs)

        return self._out_schema()

    @property
    def agent_id(self) -> AgentID:
        return self._agent_id

    @property
    def state(self) -> StateT:
        return self._state

    @property
    def out_schema(self) -> type[OutT]:
        return self._out_schema

    @abstractmethod
    async def run(
        self,
        inp_items: Any,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> Any:
        pass

    @abstractmethod
    def as_tool(
        self, tool_name: str, tool_description: str, tool_strict: bool = True
    ) -> BaseTool[BaseModel, BaseModel, CtxT]:
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeAlias, TypeVar

from pydantic import BaseModel

if TYPE_CHECKING:
    from ..run_context import CtxT, RunContextWrapper
else:
    CtxT = TypeVar("CtxT")

    class RunContextWrapper(Generic[CtxT]):
        """Runtime placeholder so RunContextWrapper[CtxT] works"""


ToolInT = TypeVar("ToolInT", bound=BaseModel, contravariant=True)  # noqa: PLC0105
ToolOutT = TypeVar("ToolOutT", bound=BaseModel, covariant=True)  # noqa: PLC0105


class ToolCall(BaseModel):
    id: str
    tool_name: str
    tool_arguments: str


class BaseTool(BaseModel, ABC, Generic[ToolInT, ToolOutT, CtxT]):
    name: str
    description: str
    in_schema: type[ToolInT]
    out_schema: type[ToolOutT]

    # Supported by OpenAI API
    strict: bool | None = None

    @abstractmethod
    async def run(
        self, inp: ToolInT, ctx: RunContextWrapper[CtxT] | None = None
    ) -> ToolOutT:
        pass

    async def __call__(
        self, ctx: RunContextWrapper[CtxT] | None = None, **kwargs: Any
    ) -> ToolOutT:
        result = await self.run(self.in_schema(**kwargs), ctx=ctx)

        return self.out_schema.model_validate(result)


ToolChoice: TypeAlias = (
    Literal["none", "auto", "required"] | BaseTool[BaseModel, BaseModel, Any]
)



import asyncio
import functools
import logging
from collections.abc import Callable, Coroutine
from time import monotonic
from typing import (
    Any,
    Generic,
    overload,
)

from tqdm.autonotebook import tqdm

from ..utils import asyncio_gather_with_pbar

from .types import (
    QueryP,
    QueryR,
    QueryT,
    RateLimDecoratorWithArgsList,
    RateLimDecoratorWithArgsSingle,
    RateLimiterState,
    RetrievalCallableList,
    RetrievalCallableSingle,
)
from .utils import partial_retrieval_callable, split_pos_args

logger = logging.getLogger(__name__)


class RateLimiterC(Generic[QueryT, QueryR]):
    def __init__(
        self,
        rpm: float,
        chunk_size: int = 1000,
        max_concurrency: int = 200,
    ):
        self._rpm = rpm
        self._max_concurrency = max_concurrency
        self._chunk_size = chunk_size

        self._lock = asyncio.Lock()
        self._state = RateLimiterState(next_request_time=0.0)
        self._semaphore = asyncio.Semaphore(self._max_concurrency)

    async def process_input(
        self,
        func_partial: Callable[[QueryT], Coroutine[Any, Any, QueryR]],
        inp: QueryT,
    ) -> QueryR:
        async with self._semaphore:
            async with self._lock:
                now = monotonic()
                if now < self._state.next_request_time:
                    await asyncio.sleep(self._state.next_request_time - now)
                self._state.next_request_time = monotonic() + 1.01 * 60.0 / self._rpm
            result = await func_partial(inp)

        return result

    async def process_inputs(
        self,
        func_partial: Callable[[QueryT], Coroutine[Any, Any, QueryR]],
        inputs: list[QueryT],
        no_tqdm: bool = False,
    ) -> list[QueryR]:
        results: list[QueryR] = []
        for i in tqdm(
            range(0, len(inputs), self._chunk_size),
            disable=no_tqdm,
            desc="Processing chunks",
        ):
            chunk = inputs[i : i + self._chunk_size]
            corouts = [
                self.process_input(func_partial=func_partial, inp=inp) for inp in chunk
            ]
            chunk_results = await asyncio.gather(*corouts)
            results.extend(chunk_results)

        return results

    @property
    def rpm(self) -> float:
        return self._rpm

    @rpm.setter
    def rpm(self, value: float) -> None:
        self._rpm = value

    @property
    def max_concurrency(self) -> int:
        return self._max_concurrency

    @property
    def chunk_size(self) -> int:
        return self._chunk_size

    @property
    def state(self) -> RateLimiterState:
        return self._state


@overload
def limit_rate(
    call: RetrievalCallableSingle[QueryT, QueryP, QueryR],
    rate_limiter: RateLimiterC[QueryT, QueryR] | None = None,
) -> RetrievalCallableSingle[QueryT, QueryP, QueryR]: ...


@overload
def limit_rate(
    call: None = None,
    rate_limiter: RateLimiterC[QueryT, QueryR] | None = None,
) -> RateLimDecoratorWithArgsSingle[QueryT, QueryP, QueryR]: ...


def limit_rate(
    call: RetrievalCallableSingle[QueryT, QueryP, QueryR] | None = None,
    rate_limiter: RateLimiterC[QueryT, QueryR] | None = None,
) -> (
    RetrievalCallableSingle[QueryT, QueryP, QueryR]
    | RateLimDecoratorWithArgsSingle[QueryT, QueryP, QueryR]
):
    if call is None:
        return functools.partial(limit_rate, rate_limiter=rate_limiter)

    @functools.wraps(call)  # type: ignore
    async def wrapper(*args: Any, **kwargs: Any) -> QueryR:
        inp: QueryT
        self_obj, inp, other_args = split_pos_args(call, args)  # type: ignore
        call_partial = partial_retrieval_callable(call, self_obj, *other_args, **kwargs)

        _rate_limiter = rate_limiter
        if _rate_limiter is None:
            _rate_limiter = getattr(self_obj, "rate_limiter", None)

        if _rate_limiter is None:
            return await call_partial(inp)
        return await _rate_limiter.process_input(func_partial=call_partial, inp=inp)

    return wrapper


@overload
def limit_rate_chunked(
    call: RetrievalCallableSingle[QueryT, QueryP, QueryR],
    rate_limiter: RateLimiterC[QueryT, QueryR] | None = None,
    no_tqdm: bool | None = None,
) -> RetrievalCallableList[QueryT, QueryP, QueryR]: ...


@overload
def limit_rate_chunked(
    call: None = None,
    rate_limiter: RateLimiterC[QueryT, QueryR] | None = None,
    no_tqdm: bool | None = None,
) -> RateLimDecoratorWithArgsList[QueryT, QueryP, QueryR]: ...


def limit_rate_chunked(
    call: RetrievalCallableSingle[QueryT, QueryP, QueryR] | None = None,
    rate_limiter: RateLimiterC[QueryT, QueryR] | None = None,
    no_tqdm: bool | None = None,
) -> (
    RetrievalCallableList[QueryT, QueryP, QueryR]
    | RateLimDecoratorWithArgsList[QueryT, QueryP, QueryR]
):
    if call is None:
        return functools.partial(
            limit_rate_chunked, rate_limiter=rate_limiter, no_tqdm=no_tqdm
        )

    @functools.wraps(call)  # type: ignore
    async def wrapper(*args: Any, **kwargs: Any) -> list[QueryR]:
        assert call is not None

        self_obj, inputs, other_args = split_pos_args(call, args)  # type: ignore
        call_partial = partial_retrieval_callable(call, self_obj, *other_args, **kwargs)

        _no_tqdm = no_tqdm
        _rate_limiter = rate_limiter
        if _no_tqdm is None:
            _no_tqdm = getattr(self_obj, "no_tqdm", False)
        if _rate_limiter is None:
            _rate_limiter = getattr(self_obj, "rate_limiter", None)

        if _rate_limiter is None:
            return await asyncio_gather_with_pbar(
                *[call_partial(inp) for inp in inputs], no_tqdm=_no_tqdm
            )
        return await _rate_limiter.process_inputs(
            func_partial=call_partial, inputs=inputs, no_tqdm=_no_tqdm
        )

    return wrapper



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Generic, cast, final

from pydantic import BaseModel

from .agent_message import AgentMessage
from .agent_message_pool import AgentMessagePool
from .comm_agent import CommunicatingAgent
from .llm import LLM, LLMSettings
from .llm_agent_state import (
    LLMAgentState,
    MakeCustomAgentState,
    SetAgentStateStrategy,
)
from .prompt_builder import (
    FormatInputArgsHandler,
    FormatSystemArgsHandler,
    PromptBuilder,
)
from .run_context import (
    CtxT,
    InteractionRecord,
    RunContextWrapper,
    SystemRunArgs,
    UserRunArgs,
)
from .tool_orchestrator import ToolCallLoopExitHandler, ToolOrchestrator
from .typing.content import ImageData
from .typing.converters import Converters
from .typing.io import (
    AgentID,
    AgentPayload,
    AgentState,
    InT,
    LLMPrompt,
    LLMPromptArgs,
    OutT,
)
from .typing.message import Conversation, Message, SystemMessage
from .typing.tool import BaseTool
from .utils import get_prompt


class LLMAgent(
    CommunicatingAgent[InT, OutT, LLMAgentState, CtxT],
    Generic[InT, OutT, CtxT],
):
    def __init__(
        self,
        agent_id: AgentID,
        *,
        # LLM
        llm: LLM[LLMSettings, Converters],
        # Input prompt template (combines user and received arguments)
        inp_prompt: LLMPrompt | None = None,
        inp_prompt_path: str | Path | None = None,
        # System prompt template
        sys_prompt: LLMPrompt | None = None,
        sys_prompt_path: str | Path | None = None,
        # System args (static args provided via RunContextWrapper)
        sys_args_schema: type[LLMPromptArgs] = LLMPromptArgs,
        # User args (static args provided via RunContextWrapper)
        usr_args_schema: type[LLMPromptArgs] = LLMPromptArgs,
        # Received args (args from another agent)
        rcv_args_schema: type[InT] = cast("type[InT]", AgentPayload),
        # Output schema
        out_schema: type[OutT] = cast("type[OutT]", AgentPayload),
        # Tools
        tools: list[BaseTool[BaseModel, BaseModel, CtxT]] | None = None,
        max_turns: int = 1000,
        react_mode: bool = False,
        # Agent state management
        set_state_strategy: SetAgentStateStrategy = "keep",
        # Multi-agent routing
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
        dynamic_routing: bool = False,
    ) -> None:
        super().__init__(
            agent_id=agent_id,
            out_schema=out_schema,
            rcv_args_schema=rcv_args_schema,
            message_pool=message_pool,
            recipient_ids=recipient_ids,
            dynamic_routing=dynamic_routing,
        )

        # Agent state
        self._state: LLMAgentState = LLMAgentState()
        self.set_state_strategy: SetAgentStateStrategy = set_state_strategy
        self._make_custom_agent_state_impl: MakeCustomAgentState | None = None

        # Tool orchestrator
        self._tool_orchestrator: ToolOrchestrator[CtxT] = ToolOrchestrator[CtxT](
            agent_id=self.agent_id,
            llm=llm,
            tools=tools,
            max_turns=max_turns,
            react_mode=react_mode,
        )

        # Prompt builder
        sys_prompt = get_prompt(prompt_text=sys_prompt, prompt_path=sys_prompt_path)
        inp_prompt = get_prompt(prompt_text=inp_prompt, prompt_path=inp_prompt_path)
        self._prompt_builder: PromptBuilder[InT, CtxT] = PromptBuilder[InT, CtxT](
            agent_id=self._agent_id,
            sys_prompt=sys_prompt,
            inp_prompt=inp_prompt,
            sys_args_schema=sys_args_schema,
            usr_args_schema=usr_args_schema,
            rcv_args_schema=rcv_args_schema,
        )

        self.no_tqdm = getattr(llm, "no_tqdm", False)

    @property
    def llm(self) -> LLM[LLMSettings, Converters]:
        return self._tool_orchestrator.llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, BaseModel, CtxT]]:
        return self._tool_orchestrator.tools

    @property
    def max_turns(self) -> int:
        return self._tool_orchestrator.max_turns

    @property
    def sys_args_schema(self) -> type[LLMPromptArgs]:
        return self._prompt_builder.sys_args_schema

    @property
    def usr_args_schema(self) -> type[LLMPromptArgs]:
        return self._prompt_builder.usr_args_schema

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.sys_prompt

    @property
    def inp_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.inp_prompt

    def format_sys_args_handler(
        self, func: FormatSystemArgsHandler[CtxT]
    ) -> FormatSystemArgsHandler[CtxT]:
        self._prompt_builder.format_sys_args_impl = func

        return func

    def format_inp_args_handler(
        self, func: FormatInputArgsHandler[InT, CtxT]
    ) -> FormatInputArgsHandler[InT, CtxT]:
        self._prompt_builder.format_inp_args_impl = func

        return func

    def make_custom_agent_state_handler(
        self, func: MakeCustomAgentState
    ) -> MakeCustomAgentState:
        self._make_custom_agent_state_impl = func

        return func

    def tool_call_loop_exit_handler(
        self, func: ToolCallLoopExitHandler[CtxT]
    ) -> ToolCallLoopExitHandler[CtxT]:
        self._tool_orchestrator.tool_call_loop_exit_impl = func

        return func

    def _parse_output(
        self,
        conversation: Conversation,
        rcv_args: InT | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> OutT:
        if self._parse_output_impl:
            return self._parse_output_impl(
                conversation=conversation, rcv_args=rcv_args, ctx=ctx, **kwargs
            )
        try:
            return self._out_schema.model_validate_json(str(conversation[-1].content))
        except Exception:
            return self._out_schema()

    @final
    async def run(
        self,
        inp_items: LLMPrompt | list[str | ImageData] | None = None,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        rcv_message: AgentMessage[InT, LLMAgentState] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **gen_kwargs: Any,  # noqa: ARG002
    ) -> AgentMessage[OutT, LLMAgentState]:
        # Get run arguments
        sys_args: SystemRunArgs = LLMPromptArgs()
        usr_args: UserRunArgs = LLMPromptArgs()
        if ctx is not None:
            run_args = ctx.run_args.get(self.agent_id)
            if run_args is not None:
                sys_args = run_args.sys
                usr_args = run_args.usr

        if entry_point:
            assert rcv_message is None, (
                "Entry point run should not have a received message"
            )
        if inp_items:
            assert rcv_message is None, (
                "There must be no received message with user inputs"
            )

        cur_state = self.state.model_copy(deep=True)

        # 1. Make system prompt (can be None)
        formatted_sys_prompt = self._prompt_builder.make_sys_prompt(
            sys_args=sys_args, ctx=ctx
        )

        # 2. Set agent state

        rcv_state = rcv_message.sender_state if rcv_message else None
        prev_mh_len = len(cur_state.message_history)

        state = LLMAgentState.from_cur_and_rcv_states(
            cur_state=cur_state,
            rcv_state=rcv_state,
            sys_prompt=formatted_sys_prompt,
            strategy=self.set_state_strategy,
            make_custom_state_impl=self._make_custom_agent_state_impl,
            ctx=ctx,
        )

        self._print_sys_msg(state=state, prev_mh_len=prev_mh_len, ctx=ctx)

        # 3. Make and add user messages (can be empty)
        user_message_batch = self._prompt_builder.make_user_messages(
            inp_items=inp_items,
            usr_args=usr_args,
            rcv_args_batch=rcv_message.payloads if rcv_message else None,
            entry_point=entry_point,
            ctx=ctx,
        )
        if user_message_batch:
            state.message_history.add_message_batch(user_message_batch)
            self._print_msgs(user_message_batch, ctx=ctx)

        if not self.tools:
            # 4. Generate messages without tools
            await self._tool_orchestrator.generate_once(agent_state=state, ctx=ctx)
        else:
            # 4. Run tool call loop (new messages are added to the message
            #    history inside the loop)
            await self._tool_orchestrator.run_loop(agent_state=state, ctx=ctx)

        # 5. Parse outputs
        batch_size = state.message_history.batch_size
        rcv_args_batch = rcv_message.payloads if rcv_message else batch_size * [None]
        val_output_batch = [
            self.out_schema.model_validate(
                self._parse_output(conversation=conv, rcv_args=rcv_args, ctx=ctx)
            )
            for conv, rcv_args in zip(
                state.message_history.batched_conversations,
                rcv_args_batch,
                strict=False,
            )
        ]

        # 6. Write interaction history to context

        if self.dynamic_routing:
            recipient_ids = self._validate_dynamic_routing(val_output_batch)
        else:
            recipient_ids = self._validate_static_routing(val_output_batch)

        if ctx:
            interaction_record = InteractionRecord(
                source_id=self.agent_id,
                recipient_ids=recipient_ids,
                inp_items=inp_items,
                sys_prompt=self.sys_prompt,
                inp_prompt=self.inp_prompt,
                sys_args=sys_args,
                usr_args=usr_args,
                rcv_args=(rcv_message.payloads if rcv_message is not None else None),
                outputs=val_output_batch,
                state=state,
            )
            ctx.interaction_history.append(
                cast(
                    "InteractionRecord[AgentPayload, AgentPayload, AgentState]",
                    interaction_record,
                )
            )

        agent_message = AgentMessage(
            payloads=val_output_batch,
            sender_id=self.agent_id,
            sender_state=state,
            recipient_ids=recipient_ids,
        )

        if not forbid_state_change:
            self._state = state

        return agent_message

    def _print_msgs(
        self,
        messages: Sequence[Message],
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        if ctx:
            ctx.printer.print_llm_messages(messages, agent_id=self.agent_id)

    def _print_sys_msg(
        self,
        state: LLMAgentState,
        prev_mh_len: int,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        if (
            len(state.message_history) == 1
            and prev_mh_len == 0
            and isinstance(state.message_history[0][0], SystemMessage)
        ):
            self._print_msgs([state.message_history[0][0]], ctx=ctx)

    # def _format_sys_args(
    #     self,
    #     sys_args: LLMPromptArgs,
    #     ctx: RunContextWrapper[CtxT] | None = None,
    # ) -> LLMFormattedSystemArgs:
    #     return self._prompt_builder.format_sys_args(sys_args=sys_args, ctx=ctx)

    # def _format_inp_args(
    #     self,
    #     usr_args: LLMPromptArgs,
    #     rcv_args: InT,
    #     ctx: RunContextWrapper[CtxT] | None = None,
    # ) -> LLMFormattedArgs:
    #     return self._prompt_builder.format_inp_args(
    #         usr_args=usr_args, rcv_args=rcv_args, ctx=ctx
    #     )

    # def _tool_call_loop_exit(
    #     self,
    #     conversation: Conversation,
    #     *,
    #     ctx: RunContextWrapper[CtxT] | None = None,
    #     **kwargs: Any,
    # ) -> bool:
    #     return self._tool_orchestrator.tool_call_loop_exit(
    #         conversation=conversation, ctx=ctx, **kwargs
    #     )



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, Generic

from ..agent_message_pool import AgentMessage, AgentMessagePool
from ..comm_agent import CommunicatingAgent
from ..run_context import CtxT, RunContextWrapper
from ..typing.io import AgentID, AgentPayload, AgentState, InT, OutT


class WorkflowAgent(
    CommunicatingAgent[InT, OutT, AgentState, CtxT],
    ABC,
    Generic[InT, OutT, CtxT],
):
    def __init__(
        self,
        agent_id: AgentID,
        subagents: Sequence[
            CommunicatingAgent[AgentPayload, AgentPayload, AgentState, CtxT]
        ],
        start_agent: CommunicatingAgent[InT, AgentPayload, AgentState, CtxT],
        end_agent: CommunicatingAgent[AgentPayload, OutT, AgentState, CtxT],
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
        dynamic_routing: bool = False,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        if not subagents:
            raise ValueError("At least one step is required")

        self.subagents = subagents

        self._start_agent = start_agent
        self._end_agent = end_agent

        super().__init__(
            agent_id=agent_id,
            out_schema=end_agent.out_schema,
            rcv_args_schema=start_agent.rcv_args_schema,
            message_pool=message_pool,
            recipient_ids=recipient_ids,
            dynamic_routing=dynamic_routing,
        )
        for subagent in subagents:
            assert not subagent.recipient_ids, (
                "Subagents must not have recipient_ids set."
            )

    @property
    def start_agent(self) -> CommunicatingAgent[InT, AgentPayload, AgentState, CtxT]:
        return self._start_agent

    @property
    def end_agent(self) -> CommunicatingAgent[AgentPayload, OutT, AgentState, CtxT]:
        return self._end_agent

    @abstractmethod
    async def run(
        self,
        inp_items: Any | None = None,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        rcv_message: AgentMessage[InT, AgentState] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **generation_kwargs: Any,
    ) -> AgentMessage[OutT, AgentState]:
        pass



import logging
from abc import abstractmethod
from collections.abc import Sequence
from typing import Any, Generic, Protocol, TypeVar, cast, final

from pydantic import BaseModel

from .agent_message import AgentMessage
from .agent_message_pool import AgentMessagePool
from .base_agent import BaseAgent
from .run_context import CtxT, RunContextWrapper
from .typing.io import AgentID, AgentPayload, AgentState, InT, OutT, StateT
from .typing.tool import BaseTool

logger = logging.getLogger(__name__)

_EH_OutT = TypeVar("_EH_OutT", bound=AgentPayload, contravariant=True)  # noqa: PLC0105
_EH_StateT = TypeVar("_EH_StateT", bound=AgentState, contravariant=True)  # noqa: PLC0105


class ExitHandler(Protocol[_EH_OutT, _EH_StateT, CtxT]):
    def __call__(
        self,
        output_message: AgentMessage[_EH_OutT, _EH_StateT],
        agent_state: _EH_StateT,
        ctx: RunContextWrapper[CtxT] | None,
    ) -> bool: ...


class CommunicatingAgent(
    BaseAgent[OutT, StateT, CtxT], Generic[InT, OutT, StateT, CtxT]
):
    def __init__(
        self,
        agent_id: AgentID,
        *,
        out_schema: type[OutT] = AgentPayload,
        rcv_args_schema: type[InT] = AgentPayload,
        recipient_ids: Sequence[AgentID] | None = None,
        message_pool: AgentMessagePool[CtxT] | None = None,
        dynamic_routing: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(agent_id=agent_id, out_schema=out_schema, **kwargs)
        self._message_pool = message_pool or AgentMessagePool()

        self._dynamic_routing = dynamic_routing

        self._is_listening = False
        self._exit_impl: ExitHandler[OutT, StateT, CtxT] | None = None

        self._rcv_args_schema = rcv_args_schema
        self.recipient_ids = recipient_ids or []

    @property
    def rcv_args_schema(self) -> type[InT]:  # type: ignore[reportInvalidTypeVarUse]
        return self._rcv_args_schema

    @property
    def dynamic_routing(self) -> bool:
        return self._dynamic_routing

    def _parse_output(
        self,
        *args: Any,
        rcv_args: InT | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> OutT:
        if self._parse_output_impl:
            return self._parse_output_impl(*args, rcv_args=rcv_args, ctx=ctx, **kwargs)

        return self._out_schema()

    def _validate_dynamic_routing(self, payloads: Sequence[OutT]) -> Sequence[AgentID]:
        assert all((p.selected_recipient_ids is not None) for p in payloads), (
            "Dynamic routing is enabled, but some payloads have no recipient IDs"
        )

        selected_recipient_ids_per_payload = [
            set(p.selected_recipient_ids or []) for p in payloads
        ]
        assert all(
            x == selected_recipient_ids_per_payload[0]
            for x in selected_recipient_ids_per_payload
        ), "All payloads must have the same recipient IDs for dynamic routing"

        assert payloads[0].selected_recipient_ids is not None
        selected_recipient_ids = payloads[0].selected_recipient_ids

        assert all(rid in self.recipient_ids for rid in selected_recipient_ids), (
            "Dynamic routing is enabled, but recipient IDs are not in "
            "the allowed agent's recipient IDs"
        )

        return selected_recipient_ids

    def _validate_static_routing(self, payloads: Sequence[OutT]) -> Sequence[AgentID]:
        assert all((p.selected_recipient_ids is None) for p in payloads), (
            "Dynamic routing is not enabled, but some payloads have recipient IDs"
        )

        return self.recipient_ids

    async def post_message(self, message: AgentMessage[OutT, StateT]) -> None:
        if self._dynamic_routing:
            self._validate_dynamic_routing(message.payloads)
        else:
            self._validate_static_routing(message.payloads)

        await self._message_pool.post(message)

    @abstractmethod
    async def run(
        self,
        inp_items: Any | None = None,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        rcv_message: AgentMessage[InT, StateT] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **kwargs: Any,
    ) -> AgentMessage[OutT, StateT]:
        pass

    async def run_and_post(
        self, ctx: RunContextWrapper[CtxT] | None = None, **run_kwargs: Any
    ) -> None:
        output_message = await self.run(
            ctx=ctx, rcv_message=None, entry_point=True, **run_kwargs
        )
        await self.post_message(output_message)

    def exit_handler(
        self, func: ExitHandler[OutT, StateT, CtxT]
    ) -> ExitHandler[OutT, StateT, CtxT]:
        self._exit_impl = func

        return func

    def _exit_condition(
        self,
        output_message: AgentMessage[OutT, StateT],
        ctx: RunContextWrapper[CtxT] | None,
    ) -> bool:
        if self._exit_impl:
            return self._exit_impl(
                output_message=output_message, agent_state=self.state, ctx=ctx
            )

        return False

    async def _message_handler(
        self,
        message: AgentMessage[AgentPayload, AgentState],
        ctx: RunContextWrapper[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        rcv_message = cast("AgentMessage[InT, StateT]", message)
        out_message = await self.run(ctx=ctx, rcv_message=rcv_message, **run_kwargs)

        if self._exit_condition(output_message=out_message, ctx=ctx):
            await self._message_pool.stop_all()
            return

        if self.recipient_ids:
            await self.post_message(out_message)

    @property
    def is_listening(self) -> bool:
        return self._is_listening

    async def start_listening(
        self, ctx: RunContextWrapper[CtxT] | None = None, **run_kwargs: Any
    ) -> None:
        if self._is_listening:
            return

        self._is_listening = True
        self._message_pool.register_message_handler(
            agent_id=self.agent_id,
            handler=self._message_handler,
            ctx=ctx,
            **run_kwargs,
        )

    async def stop_listening(self) -> None:
        self._is_listening = False
        await self._message_pool.unregister_message_handler(self.agent_id)

    @final
    def as_tool(
        self, tool_name: str, tool_description: str, tool_strict: bool = True
    ) -> BaseTool[BaseModel, BaseModel, CtxT]:
        # assert self.state.batch_size == 1, (
        #     "Using agents as tools is only supported for batch size 1"
        # )

        agent_instance = self

        class AgentTool(BaseTool[BaseModel, BaseModel, Any]):
            name: str = tool_name
            description: str = tool_description
            in_schema: type[BaseModel] = agent_instance.rcv_args_schema
            out_schema: type[BaseModel] = agent_instance.out_schema

            strict: bool | None = tool_strict

            async def run(
                self,
                inp: BaseModel,
                ctx: RunContextWrapper[CtxT] | None = None,
            ) -> OutT:
                rcv_args = agent_instance.rcv_args_schema.model_validate(inp)
                rcv_message = AgentMessage(  # type: ignore[arg-type]
                    payloads=[rcv_args],
                    sender_id="<tool_user>",
                    recipient_ids=[agent_instance.agent_id],
                )

                agent_result = await agent_instance.run(
                    rcv_message=rcv_message,  # type: ignore[arg-type]
                    entry_point=False,
                    forbid_state_change=True,
                    ctx=ctx,
                )

                return agent_result.payloads[0]

        return AgentTool()



import inspect
from collections.abc import Callable, Coroutine, Sequence
from typing import (
    Any,
)

from .types import (
    QueryP,
    QueryR,
    QueryT,
    RetrievalCallableList,
    RetrievalCallableSingle,
)


def is_bound_method(func: Callable[..., Any], self_candidate: Any) -> bool:
    return (inspect.ismethod(func) and (func.__self__ is self_candidate)) or hasattr(self_candidate, func.__name__)


def split_pos_args(
    call: (RetrievalCallableSingle[QueryT, QueryP, QueryR] | RetrievalCallableList[QueryT, QueryP, QueryR]),
    args: Sequence[Any],
) -> tuple[Any | None, QueryT | list[QueryT], Sequence[Any]]:
    if not args:
        raise ValueError("No positional arguments passed.")
    maybe_self = args[0]
    if is_bound_method(call, maybe_self):
        # Case: Bound instance method with signature (self, inp, *rest)
        if len(args) < 2:
            raise ValueError(
                "Must pass at least `self` and an input (or a list of inputs) " + "for a bound instance method."
            )
        return maybe_self, args[1], args[2:]
    # Case: Standalone function with signature (inp, *rest)
    if not args:
        raise ValueError("Must pass an input (or a list of inputs) " + "for a standalone function.")
    return None, args[0], args[1:]


def partial_retrieval_callable(
    call: Callable[..., Coroutine[Any, Any, QueryR]],
    self_obj: Any,
    *args: QueryP.args,
    **kwargs: QueryP.kwargs,
) -> Callable[[QueryT], Coroutine[Any, Any, QueryR]]:
    async def wrapper(inp: QueryT) -> QueryR:
        if self_obj is not None:
            # `call` is a method
            return await call(self_obj, inp, *args, **kwargs)
        # `call` is a function
        return await call(inp, *args, **kwargs)

    return wrapper


def expected_exec_time_from_max_concurrency_and_rpm(rpm: float, max_concurrency: int) -> float:
    return 60.0 / (rpm / max_concurrency)
